static char *recvmsg_id = "@(#) $Id: recvmsg.c,v 1.4 1999/11/16 06:00:51 kouno Exp $";

/* 
 *  Project: IPMSG
 *  Author & Copyright:  KOHNO TAKASHI kouno@catnet.ne.jp
 *  Mail: kouno@catnet.ne.jp
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <limits.h>
#include <netdb.h>
#include <signal.h>
#include <pwd.h>
#include <errno.h>
#include <time.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

#if defined(DEBUG)
#define line	fprintf(stderr, "Line %d\n", __LINE__);fflush(stderr);fflush(stdout)
#define d_printf printf
#else
#define debug
#define d_printf dummy_function
#endif

#define MAILMSG_VERSION "0.51"

#define IPMSG_VERSION		0x001
#define IPMSG_PORT			0x979
/* #define IPMSG_PORT 2000 */
#define IPMSG_REVERSEICON	0x100

/*  command  */
#define IPMSG_NOOPERATION 0x00000000UL

#define IPMSG_BR_ENTRY 0x00000001UL
#define IPMSG_BR_EXIT 0x00000002UL
#define IPMSG_ANSENTRY 0x00000003UL
#define IPMSG_BR_ABSENCE 0x00000004UL

#define IPMSG_BR_ISGETLIST 0x00000010UL
#define IPMSG_OKGETLIST 0x00000011UL
#define IPMSG_GETLIST 0x00000012UL
#define IPMSG_ANSLIST 0x00000013UL
#define IPMSG_BR_ISGETLIST2 0x00000018UL

#define IPMSG_SENDMSG 0x00000020UL
#define IPMSG_RECVMSG 0x00000021UL
#define IPMSG_READMSG 0x00000030UL
#define IPMSG_DELMSG 0x00000031UL

#define IPMSG_GETINFO 0x00000040UL
#define IPMSG_SENDINFO 0x00000041UL

#define IPMSG_GETABSENCEINFO 0x00000050UL
#define IPMSG_SENDABSENCEINFO 0x00000051UL

/*  option for all command  */
#define IPMSG_ABSENCEOPT 0x00000100UL
#define IPMSG_SERVEROPT 0x00000200UL
#define IPMSG_DIALUPOPT 0x00010000UL

/*  option for send command  */
#define IPMSG_SENDCHECKOPT 0x00000100UL
#define IPMSG_SECRETOPT 0x00000200UL
#define IPMSG_BROADCASTOPT 0x00000400UL
#define IPMSG_MULTICASTOPT 0x00000800UL
#define IPMSG_NOPOPUPOPT 0x00001000UL
#define IPMSG_AUTORETOPT 0x00002000UL
#define IPMSG_RETRYOPT 0x00004000UL
#define IPMSG_PASSWORDOPT 0x00008000UL
#define IPMSG_NOLOGOPT 0x00020000UL
#define IPMSG_NEWMUTIOPT 0x00040000UL
#define IPMSG_NOADDLISTOPT 0x00080000UL

#define HOSTLIST_DELIMIT "\a"
#define HOSTLIST_DUMMY "\b"

/*  end of IP Messenger Communication Protocol version 1.0 define  */

#define UDP_BUF 8192
#define MAXBUFLEN 0x2ff
#define BROADCAST "DEFAULT"
#define TIPMSGDIR "tipmsg"
#define RECVMSGBC "recvmsgbc"
#define RECVMSGBC_BWC ".recvmsgbc"
#define DEFAULT_USERLIST "userlist"
#define ENV_HOME "HOME"
#define DEFAULT_SAVEFILE "save"
#define DEFAULT_LOGFILE "recvmsg.log"
#define MAXBCNUM 0x100
#define MAXBCLEN 16
#define MAXSTPACKET 100
#define MAX_RETRY 1
#define RETRY_WAIT 2
#define DEFAULT_RECENTFILE "recent_recvfrom"
#define SAVE_FILE_NUM_MAX "3"

#define GET_MODE(command)	(command & 0x000000ff)
#define GET_OPT(command)	(command & 0x0000ff00)

typedef struct {
    int   format_ver;
    int   packet_no;
    char  msg_user[MAXBUFLEN];
    char  msg_host[MAXBUFLEN];
    int   command;
    int   group_offset;
    char  extend[UDP_BUF];
} MSG;

typedef struct USERLIST_TMP {
    struct USERLIST_TMP* next;
    struct USERLIST_TMP* prev;
	struct sockaddr_in	saddr;

    MSG msg; /* Must last location */
} USERLIST;

typedef struct {
    USERLIST* first;
    USERLIST* last;
} USERLIST_MNG;

int sender(int sd, char *buf, long addr, int command);
int recver(int sd);
long hosttoaddr(char *target);
int decom_msg(MSG *msg, char *buf);
void interrupt_exit();
void print_usage(void);
int print_message(int sd, MSG *msg, struct sockaddr_in* psaddr);
int print_message2(FILE* fp, MSG* msg);

int append_userlist(USERLIST_MNG* ulm, USERLIST* ul);
int delete_userlist(USERLIST_MNG* ulm, USERLIST* ul);
void destroy_userlist(USERLIST_MNG* ulm);
USERLIST* search_userlist(USERLIST* uls, USERLIST* ul, int flag);
USERLIST* search_userlist2(USERLIST* uls, long addr, int flag);
int flush_userlist(char* path, USERLIST_MNG* ulm);

void dummy_function(char *dummy,...);

int g_sd;
int g_ipmsg_port = IPMSG_PORT;
int g_broadcast_num = 0;
char g_broadcast[MAXBCNUM][MAXBCLEN];
int g_get_packet_num = 0;
int g_storage_packet[MAXSTPACKET];
int g_nflag = 1;
int g_timer_exit = 0;
char g_save_file_master[PATH_MAX] = "";
char g_save_file_number[PATH_MAX] = "";
char g_log_file[PATH_MAX] = "";
char g_userlist_file[PATH_MAX] = "";
int g_exit = 0;
char g_name[MAXBCNUM] = "";
char g_group[MAXBCNUM] = "";
USERLIST_MNG g_userlist_mng = {NULL, NULL};
char g_recent_recvfrom_file[PATH_MAX] = "";

int main(int argc, char *argv[])
{
	int	sd;
	char buf[UDP_BUF];
	struct hostent	*host;
	long addr;
    char* ptr;
    int lp;
    FILE* fp;
    char bcfile[PATH_MAX];
    int flag=1;
    char* p;
    int ret = 0;

    memset(g_storage_packet, 0, MAXSTPACKET);

    sprintf(buf, "%s/%s", getenv(ENV_HOME), TIPMSGDIR);
    mkdir(buf, 0755);

	/* ̉ - nCt1ɂ1̃o[W*/
	for( ;argc > 1 && argv[1][0] == '-';argv++,argc-- ){
		switch( argv[1][1] ){
			case '?':
			case 'h':
                print_usage();
				exit(0);
			case 'p':
			case 'P':
				g_ipmsg_port=strtol(&argv[1][2], &ptr, 0);
				fprintf(stderr, "PORT %d(%s)\n", g_ipmsg_port, &argv[1][0]);
				if (g_ipmsg_port==0)
				  g_ipmsg_port=IPMSG_PORT;
				break;
			case 'n':
			case 'N':
				g_nflag = 0;
				break;
			case 'e':
			case 'E':
				g_timer_exit = strtol(&argv[1][2], &ptr, 0);
				fprintf(stderr, "TIMER EXIT %d(%s)\n", g_timer_exit, &argv[1][0]);
				break;
            case 'f':
            case 'F':
                if (argv[1][2] != '\0') {
                    strcpy(g_save_file_master, &argv[1][2]);
                } else {
                    sprintf(g_save_file_master, "%s/%s/%s", getenv(ENV_HOME), TIPMSGDIR, DEFAULT_SAVEFILE);
                    mkdir(g_save_file_master, 0700);
                }
                d_printf("Enable file saving(%s)\n", g_save_file_master);
                break;
            case 'l':
            case 'L':
                if (argv[1][2] != '\0') {
                    strcpy(g_log_file, &argv[1][2]);
                } else {
                    sprintf(g_log_file, "%s/%s/%s", getenv(ENV_HOME), TIPMSGDIR, DEFAULT_LOGFILE);
                }
                d_printf("Enable message loging(%s)\n", g_log_file);
                break;
            case 'u':
            case 'U':
                if (argv[1][2] != '\0') {
                    strcpy(g_userlist_file, &argv[1][2]);
                } else {
                    sprintf(g_userlist_file, "%s/%s/%s", getenv(ENV_HOME), TIPMSGDIR, DEFAULT_USERLIST);
                }
                d_printf("Enable Userlist(%s)\n", g_userlist_file);
                break;
            case 'r':
            case 'R':
                if (argv[1][2] != '\0') {
                    strcpy(g_recent_recvfrom_file, &argv[1][2]);
                } else {
                    sprintf(g_recent_recvfrom_file, "%s/%s/%s", 
                            getenv(ENV_HOME), TIPMSGDIR, DEFAULT_RECENTFILE);
                }
                d_printf("Enable Recent recvfrom.(%s)\n", g_recent_recvfrom_file);
                break;
            case 's':
            case 'S':
                /* pɗ\ */
                break;
			default:
				break;
            }
    }

    if (argc>1){
        ptr = argv[1];
    } else {
        ptr = bcfile;
        sprintf(bcfile, "%s/%s/%s", getenv(ENV_HOME), TIPMSGDIR, RECVMSGBC);
        if (access(ptr, F_OK) != 0){
            fprintf(stderr, "Waring: Not found [%s]. Try Using old style file.\n", ptr);
            sprintf(bcfile, "%s/%s", getenv(ENV_HOME), RECVMSGBC_BWC);
        }
    }
    
    if (fp = fopen(ptr, "r")){
        for (g_broadcast_num=0; 
             fgets(g_broadcast[g_broadcast_num], MAXBCLEN, fp) && g_broadcast_num < MAXBCNUM;
             g_broadcast_num++){
            for (p = g_broadcast[g_broadcast_num]; *p; p++){
                if (*p == '\n' || *p == '\r'){
                    *p = '\0';
                    break;
                }
            }
            if (g_broadcast[g_broadcast_num][0] == '#'){
                g_broadcast_num--;
                continue;
            } else if (g_broadcast[g_broadcast_num][0] == '!' ||
                       g_broadcast[g_broadcast_num][0] == 'N'){
                for (p=&g_broadcast[g_broadcast_num][1]; *p;p++){
                    if (*p != '\t' && *p != ' '){
                        strcpy(g_name, p);
                        break;
                    }
                }
                g_broadcast_num--;
                continue;
            } else if (g_broadcast[g_broadcast_num][0] == '%' ||
                       g_broadcast[g_broadcast_num][0] == 'G'){
                for (p=&g_broadcast[g_broadcast_num][1]; *p;p++){
                    if (*p != '\t' && *p != ' '){
                        strcpy(g_group, p);
                        break;
                    }
                }
                g_broadcast_num--;
                continue;
            }
            g_broadcast[g_broadcast_num][MAXBCLEN-1] = '\0';
        }
        fclose(fp);
    } else {
        fprintf(stderr, "Can't open %s file : ", ptr);
        perror(NULL);
        g_broadcast_num = 1;
        strcpy(g_broadcast[0], BROADCAST);
    }

	if (!g_name[0]){
        strcpy(g_name, getpwuid(getuid())->pw_name);
    }


	if ((sd = socket(AF_INET, SOCK_DGRAM, 0)) ==-1){
		perror("Can't create socket\n");
	}
	
    if (setsockopt(sd, SOL_SOCKET, SO_BROADCAST, (char*)&flag, sizeof(int)) == -1) {
        perror("setsockopt fail");
    }

    /* Ctl-C */
    g_sd = sd;
    signal(SIGINT, interrupt_exit);
    /* Timer */
    if (g_timer_exit){
        struct itimerval value;
        signal(SIGALRM, interrupt_exit);

        value.it_interval.tv_sec = g_timer_exit;
        value.it_interval.tv_usec = 0;
        value.it_value.tv_sec = g_timer_exit;
        value.it_value.tv_usec = 0;
        setitimer(ITIMER_REAL, &value, NULL);
    }

	/* M҂ */
	if (!recver(sd)){
		ret = 1;
	}
    if (g_userlist_file[0]){
        destroy_userlist(&g_userlist_mng);
    }

    for (lp=0; lp<g_broadcast_num; lp++){
        sender(sd, NULL, hosttoaddr(g_broadcast[lp]), IPMSG_BR_EXIT);
        sender(sd, NULL, hosttoaddr(g_broadcast[lp]), IPMSG_BR_EXIT);
        sender(sd, NULL, hosttoaddr(g_broadcast[lp]), IPMSG_BR_EXIT);
    }
	close( sd );
	return ret;
}

/* Sender */
int sender( int sd, char *buf, long addr, int command)
{
	char full_buf[UDP_BUF];
    char hostname[MAXBUFLEN];
	static long packet_no = 0;
	struct sockaddr_in	saddr;
	struct hostent		*host;
    char* p;

    if (!packet_no){
        packet_no = time(&packet_no) + ((long) getpid() << 16);
    } else {
        packet_no++;
    }

	memset(&saddr, 0, sizeof(saddr));
	saddr.sin_family		= AF_INET;
	saddr.sin_port			= htons(g_ipmsg_port);
	saddr.sin_addr.s_addr	= addr;

	if (gethostname(hostname, MAXBUFLEN) == -1){
		strcpy(hostname,"Unknow host");
	}
    if (p = strchr(hostname, '.')){
        *p = '\0';
    }
    
	/* Make Message */
	sprintf(full_buf, "%d:%d:%s:%s:%d:%s", IPMSG_VERSION, packet_no, getpwuid(getuid())->pw_name, hostname, command, buf ? buf : "");
    
	/* Send */
	if (sendto(sd, full_buf, strlen(full_buf) +1, 0, (struct sockaddr *)&saddr, sizeof(saddr)) == -1){
		perror("sendto err");
		return	0;
	}
    
	return 1;
}

int entry_sender( int sd, char* name, char* group, long addr, int command)
{
	char buf[UDP_BUF];
    char hostname[MAXBUFLEN];
	static int packet_no;
	struct sockaddr_in	saddr;
	struct hostent		*host;
    char* p;
    int len;

	memset(&saddr, 0, sizeof(saddr));
	saddr.sin_family		= AF_INET;
	saddr.sin_port			= htons(g_ipmsg_port);
	saddr.sin_addr.s_addr	= addr;

	if (gethostname(hostname, MAXBUFLEN) == -1){
		strcpy(hostname,"Unknow host");
	}
    if (p = strchr(hostname, '.')){
        *p = '\0';
    }

	/* Make Message */
	sprintf(buf, "%d:%d:%s:%s:%d:%s%c%s%c", IPMSG_VERSION, packet_no++, getpwuid(getuid())->pw_name, hostname, command, name ? name : "", '\0', group ? group : "", '\0');

	/* Send */
    len = strlen(buf);
    len += strlen(buf + len + 1) + 1;
	if (sendto(sd, buf, len, 0, (struct sockaddr *)&saddr, sizeof(saddr)) == -1){
		perror("sendto err");
		return	0;
	}

	return 1;
}

/* zXg -> AhX ϊ */
long hosttoaddr(char *target)
{
	struct sockaddr_in	saddr;
    char hostname[MAXBUFLEN];
	long addr;
	struct hostent *host;

    if (strcmp(target, BROADCAST) == 0){
        return ~0;
    }

	if ((addr = inet_addr(target)) == -1){
		if ((host = (struct hostent *)gethostbyname(target)) != NULL){
			addr = *(long *)host->h_addr;
		} else {
            fprintf(stderr, "invalid hostname %s\n", target);
			perror(NULL);
			return 0;
		}
	}
    
	return addr;
}

/* Receive Message function */
int recver(int sd)
{
	char buf[UDP_BUF];
	char str[UDP_BUF];
/*	char message[UDP_BUF]=""; */
    char recent_recvfrom[PATH_MAX];
	struct sockaddr_in	saddr;
	MSG* pmsg;
	int flag=1;
	int	size;
	int lp;
    fd_set rfds;
    struct timeval tv;
    int retval;
    USERLIST* ul;
    int ff=0; /* Flush userlist flag */
    FILE* fp;

	memset(&buf, 0, sizeof(buf));
	size=sizeof(saddr);

	/* bZ[W{̗p̗̈m */
	if (!(pmsg = (MSG*)calloc(1, sizeof(MSG)))){
		perror("Message memory allocate error!\n");
		goto ERR0;
	}
	if (!(ul = (USERLIST*)calloc(1, sizeof(USERLIST)))){
		perror("Message memory allocate error!\n");
        goto ERR1;
	}
    pmsg->extend[0] = '\0';

	/* sockaddr_in\̂̐ݒ */
	memset((char *)&saddr, 0, size );
	saddr.sin_family		= AF_INET;
	saddr.sin_port			= htons(g_ipmsg_port);
	saddr.sin_addr.s_addr	= htonl(INADDR_ANY);

	/* \PbgoCh */
    for (lp=0; 1; lp++){
        if (bind(sd, (struct sockaddr *)&saddr, sizeof(saddr)) == -1){
            if (errno == EADDRINUSE && lp<MAX_RETRY){
                sleep(RETRY_WAIT);
            } else {
                fprintf(stderr, "Bind error %d port\n", g_ipmsg_port);
                perror(NULL);
                goto ERR2;
            }
        } else {
            break;
        }
    }

	/* Gg[ʒm */
    for (lp=0; lp<g_broadcast_num; lp++){
        if (!entry_sender(sd, g_name, g_group, hosttoaddr(g_broadcast[lp]), IPMSG_BR_ENTRY)){
            goto ERR2;
        }
    }

	while(flag){
        if (g_exit){
            flag = 0;
        }
        FD_ZERO(&rfds);
        FD_SET(sd, &rfds);
        tv.tv_sec = 1;
        tv.tv_usec = 0;
        
        retval = select(sd+1, &rfds, NULL, NULL, &tv);
        if (retval == -1){
            perror("select fail");
            flush_userlist(g_userlist_file, &g_userlist_mng);
            goto ERR2;
        } else if (!retval){
            if (ff){
                flush_userlist(g_userlist_file, &g_userlist_mng);
                ff=0;
            }
            continue;
        }
        
        /* tJn */
		if ((retval = recvfrom(sd, buf, UDP_BUF, 0, (struct sockaddr *)&saddr, &size)) == -1){
            perror("recv error");
            goto ERR2;
		}

		if ( !decom_msg(pmsg, buf) ){
			continue;
		}
		
        /* Mobt@̃[NA(e͊ msg \̂ɑ݂) */
        memset(buf, 0, retval);

		/* R}h̔ */
		switch(GET_MODE(pmsg->command)){
        case IPMSG_BR_ENTRY:
            if ( !entry_sender(sd, g_name, g_group, saddr.sin_addr.s_addr, IPMSG_ANSENTRY)){
                goto ERR2;
            }
            /* ʂ蔲` */
        case IPMSG_ANSENTRY:
        	if (g_nflag){
	            fprintf(stdout, "Entry (%s@%s%s%s[%s])\n", pmsg->extend[0] ? pmsg->extend : pmsg->msg_user, pmsg->extend + pmsg->group_offset, pmsg->extend[pmsg->group_offset] ? "/" : "", pmsg->msg_host, inet_ntoa(saddr.sin_addr));
	        }
            ul->msg = *pmsg;
            ul->saddr = saddr;
            if (g_userlist_file[0]){
                append_userlist(&g_userlist_mng, ul);
            }
            ff=1;
            break;
        case IPMSG_BR_EXIT:
        	if (g_nflag){
	            fprintf(stdout, "Exit (%s@%s[%s])\n", pmsg->msg_user, pmsg->msg_host, inet_ntoa(saddr.sin_addr));
	        }
            ul->msg = *pmsg;
            ul->saddr = saddr;
            if (g_userlist_file[0]){
                delete_userlist(&g_userlist_mng, ul);
            }
            ff=1;
            break;
        case IPMSG_RECVMSG:
            fprintf(stdout, "%s@%s 烁bZ[W̎MmF󂯎܂\n" , pmsg->msg_user, pmsg->msg_host);
            break;
        case IPMSG_READMSG:
            fprintf(stdout, "%s@%s 烁bZ[W̊JmF󂯂Ƃ܂\n", pmsg->msg_user, pmsg->msg_host);
            break;
        case IPMSG_SENDMSG:
            if (!print_message(sd, pmsg, &saddr)){
                goto ERR2;
            }
            if (g_recent_recvfrom_file[0]){
                if (ul = search_userlist2(g_userlist_mng.first, saddr.sin_addr.s_addr, 0)){
                    sprintf(recent_recvfrom, "%s@%s%s%s[%s](%s:%s)", 
                            ul->msg.extend[0] ? ul->msg.extend : ul->msg.msg_user, 
                            ul->msg.extend + ul->msg.group_offset, 
                            ul->msg.extend[ul->msg.group_offset] ? "/" : "", 
                            ul->msg.msg_host, inet_ntoa(ul->saddr.sin_addr), 
                            ul->msg.msg_user, ul->msg.msg_host);
                } else {
                    d_printf("Can't find message user in userlist.\n");
                    sprintf(recent_recvfrom, "%s@%s[%s](%s:%s)", 
                            pmsg->msg_user, 
                            ul->msg.msg_host, inet_ntoa(ul->saddr.sin_addr), 
                            ul->msg.msg_user, ul->msg.msg_host);
                }
                if (fp = fopen(g_recent_recvfrom_file, "w")){
                    fprintf(fp, recent_recvfrom);
                    fclose(fp);
                    d_printf("Set recent recvfrom.(%s)\n", recent_recvfrom);
                } else {
                    fprintf(stderr, "Can't open recent recvfrom file.(%s)\n", g_recent_recvfrom_file);
                }
            }
            break;
		}
        fflush(stdout);
	}
	return 1;

 ERR2:
    free(ul);
 ERR1:
    free(pmsg);
 ERR0:
    return 0;
}

int decom_msg( MSG *msg, char *buf )
{
	char *str;
	
	/* Get format version */
	if ((str=strchr( buf,':')) == NULL){
		fprintf(stderr, "Can't find format version\n" );
		return 0;
	}
	str[0]='\0';
	if ((msg->format_ver=atoi(buf)) != IPMSG_VERSION){
		return 0;
	}
	buf=str+1;
	
	/* Get Packet No */
	if ((str=strchr(buf, ':')) == NULL){
		fprintf(stderr, "Can't find Packet\n");
		return 0;
	}
	str[0]='\0';
	msg->packet_no=atoi( buf ) ;
	buf=str+1;
	
	/* Get Sender Name */
	if ((str=strchr(buf, ':')) == NULL){
		fprintf(stderr, "Can't find Sender Name\n" );
		return 0;
	}
	str[0]='\0';
	strcpy( msg->msg_user, buf );
	buf=str+1;
	
	/* Get Sendhost Name */
	if ((str=strchr(buf,':')) == NULL){
		fprintf(stderr, "Can't find Sendhost Name\n" );
		return 0;
	}
	str[0]='\0';
	strcpy( msg->msg_host, buf );
	buf=str+1;
	
	/* Get command */
	if ((str=strchr(buf,':')) == NULL){
		fprintf(stderr, "Can't find command\n" );
		return 0;
	}
	str[0]='\0';
	msg->command=atoi(buf) ;
	buf=str+1;
	
	/* Get Extension */
	sprintf(msg->extend, "%s%c%s", buf, '\0', buf + strlen(buf)+1);
    msg->group_offset = strlen(msg->extend)+1;
	
	return 1;
}

/* void interrupt_exit(int sig, int code, struct sigcontext *scp, char *addr)*/
void interrupt_exit()
{
    int lp, lp2;
    
    signal(SIGINT, SIG_IGN);
    for (lp=0; lp<g_broadcast_num; lp++){
        sender(g_sd, NULL, hosttoaddr(g_broadcast[lp]), IPMSG_BR_EXIT);
        sender(g_sd, NULL, hosttoaddr(g_broadcast[lp]), IPMSG_BR_EXIT);
        sender(g_sd, NULL, hosttoaddr(g_broadcast[lp]), IPMSG_BR_EXIT);
    }
    
    g_exit = 1;
    signal(SIGINT, SIG_DFL);
}

void print_usage()
{
    printf("resvmsg ver %s - IPMSG recever tool\n", MAILMSG_VERSION);
    printf("Usage: recvmsg -[npefh] [Broadcast file]\n");
    printf("        -n : Contain [ENTRY]and[EXIT] message.\n");
    printf("        -p : Designate port number.\n");
    printf("        -e : Designate exit timer.\n");
    printf("        -f : Designate file save.\n");
    printf("        -h : Print this message.\n");
    printf("        -u : Using userlist.\n");
    printf("Ex. recvmsg -n -p2000 -e5 -f/var/log/recvmsg -u/tmp/userlist ~/.recvmsgbc\n");
}

int print_message(int sd, MSG* msg, struct sockaddr_in* psaddr)
{
	char buf[UDP_BUF];
    int lp;
    FILE* fp;
    static int count = 0;
    char linkbuf[PATH_MAX];
    static char master[PATH_MAX]="";
    char *endp;
    char ans[3];
    char *startp;

    fp = stdout;
    /* M`FbN */
    if (GET_OPT(msg->command) & IPMSG_SENDCHECKOPT){
        sprintf(buf, "%d", msg->packet_no);
        if (!sender(sd, buf, psaddr->sin_addr.s_addr, IPMSG_RECVMSG)){
            return 0;
        }
    }
    if (GET_OPT(msg->command) & IPMSG_SECRETOPT){
        sprintf(buf, "%d", msg->packet_no);
        if (!sender(sd, buf, psaddr->sin_addr.s_addr, IPMSG_READMSG)){
            return 0;
        }
    }
    
    fprintf(stdout, "\a\a\n" );
    print_message2(stdout, msg);

    if (g_save_file_master[0]){
        if (!master[0]){
            sprintf(master, "%s/0", g_save_file_master);
        }
        if (!count){
            if (readlink(master, linkbuf, PATH_MAX)){
                if (startp = strrchr(linkbuf, '/')){
                    d_printf("link to %s\n", startp);
                    count = strtol(startp+1, &endp, 10);
                }
            }
        }
        count++;
        sprintf(g_save_file_number, "%s/%0" SAVE_FILE_NUM_MAX "d", g_save_file_master, count);
        if ((fp = fopen(g_save_file_number, "w"))){
            print_message2(fp, msg);
        
            fclose(fp);
            unlink(master);
            symlink(g_save_file_number, master);
        } else {
            fprintf(stderr, "Can't open save file(%s)", g_save_file_number);
            perror(NULL);
        }
    }

    if (g_log_file[0]){
        if ((fp = fopen(g_log_file, "a"))){
            print_message2(fp, msg);
        
            fclose(fp);
            unlink(master);
            symlink(g_save_file_number, master);
        } else {
            fprintf(stderr, "Can't open log file(%s)", g_save_file_number);
            perror(NULL);
        }
    }

    return (strncmp(msg->extend, "quit", 4) == 0) ? 0 : 1;
}

int print_message2(FILE* fp, MSG* msg)
{
    time_t now_time;
    time( &now_time );

    fprintf(fp, "----------------------------------------------\n" );
    
    /* u[hLXg */
    if (GET_OPT(msg->command) & IPMSG_BROADCASTOPT){
        fprintf(fp, "      Broadcast from %s ( %s )\n", 
                msg->msg_user, msg->msg_host );
    } else {				
        /* }`LXg */
        if (GET_OPT(msg->command) & IPMSG_MULTICASTOPT){
            fprintf(fp, "      Multicast from %s ( %s )\n", 
                    msg->msg_user, msg->msg_host );
        } else {
            fprintf(fp, "      Message from %s ( %s )\n", 
                    msg->msg_user, msg->msg_host );
        }
    }
    fprintf(fp, "      at %s", ctime( &now_time )  );
    fprintf(fp, "----------------------------------------------\n" );
    fprintf(fp, "%s\n", msg->extend);
    fprintf(fp, "----------------Message End-------------------\n\n" );

    return 1;
}

int append_userlist(USERLIST_MNG* ulm /* Userlist Last */, USERLIST* ul)
{
    USERLIST* ult; /* Userlist Target*/
    if (ult = search_userlist(ulm->last, ul, 1)){
        /* User exist */
        return 1;
    }

    d_printf("Append %s@%s\n", ul->msg.msg_user, ul->msg.msg_host);
    if (!(ult = (USERLIST*)malloc(sizeof(USERLIST)))){
        perror("Can't allocate memory.");
        return 0;
    }

    *ult = *ul;
    ult->next = NULL;
    ult->prev = ulm->last;

    if (ulm->last){
        ulm->last->next = ult;
        ulm->last = ult;
    } else {
        d_printf("Insert first\n");
        ulm->first = ulm->last = ult;
    }
    
    return 1;
}

int delete_userlist(USERLIST_MNG* ulm /* Userlist First */, USERLIST* ul)
{
    USERLIST* ult; /* Userlist Target*/
    if (!(ult = search_userlist(ulm->first, ul, 0))){
        /* User exist */
        return 0;
    }

    if (ult->prev){
        ult->prev->next = ult->next;
    } else {
        ulm->first = ult->next;
    }
    if (ult->next){
        ult->next->prev = ult->prev;
    } else {
        ulm->last = ult->prev;
    }

    free(ult);
    return 1;
}

void destroy_userlist(USERLIST_MNG* ulm)
{
    USERLIST* ult; /* Userlist Target*/
    USERLIST* ulo;

    d_printf("Destroy_userlist Start!\n");
    while(ult = ulm->first){
        d_printf("Delete Userlist[%s]\n", ult->msg.msg_user);
        ulm->first = ult->next;
        free(ult);
    }
    d_printf("Destroy_userlist E n d!\n");
}

USERLIST* search_userlist(USERLIST* uls, USERLIST* ul, int flag /* Serach with backword*/)
{
    if (!ul){
        return NULL;
    }

    while(uls){
/*
        d_printf("Target user(%s)\n", inet_ntoa(uls->saddr.sin_addr));
        d_printf("(%08x) ?? (%08x) : %08x <-> %08x : %d\n", 
                 uls->saddr.sin_addr.s_addr, 
                 ul->saddr.sin_addr.s_addr,
                 uls->prev, uls->next,
                 flag);
*/
        if (uls->saddr.sin_addr.s_addr == ul->saddr.sin_addr.s_addr){
            d_printf("Search user(%s)\n", inet_ntoa(ul->saddr.sin_addr));
            return uls;
        }
        if (!flag){
            uls = uls->next;
        } else {
            uls = uls->prev;
        }
    }
    return NULL;
}

USERLIST* search_userlist2(USERLIST* uls, long addr, int flag /* Serach with backword*/)
{
    while(uls){
        if (uls->saddr.sin_addr.s_addr == addr){
            d_printf("Search user(%s)\n", inet_ntoa(addr));
            return uls;
        }
        if (!flag){
            uls = uls->next;
        } else {
            uls = uls->prev;
        }
    }
    return NULL;
}

int flush_userlist(char* path, USERLIST_MNG* ulm)
{
    USERLIST* ul;
    FILE* fp;

    if (!path[0] || !ulm->first){
        return 0;
    }

    d_printf("flush_userlist\n");

    if (!(fp = fopen(path, "w"))){
        fprintf(stderr, "Can't open userlist file(%s):", path);
        perror(NULL);
        return 0;
    }
    
    for (ul=ulm->first; ul; ul=ul->next){
        fprintf(fp, "%s@%s%s%s[%s](%s:%s)\n", 
                ul->msg.extend[0] ? ul->msg.extend : ul->msg.msg_user, 
                ul->msg.extend + ul->msg.group_offset, 
                ul->msg.extend[ul->msg.group_offset] ? "/" : "", 
                ul->msg.msg_host, inet_ntoa(ul->saddr.sin_addr), 
                ul->msg.msg_user, ul->msg.msg_host);
    }
    fclose(fp);
    return 1;
}

void dummy_function(char *dummy,...)
{
}
